{-
 -
 - APCC
 - (C) Ivana Kojovic, DTU 2012
 - All Rights Reserved.
 -
 - 22nd April - changes in subTerms - call for Asycrypt key has to be different than mapping
 -              msgPattern -  kick out condition for the constant, that will be checked by first condition of the theorem1
 -              error instead of False in the theorem1 condition #1
 -              error instead of False in the checkUnify, report
 -
 -
 -}


-- | Module @FormatTypeSafe@ has all needed to make subterms of single message, array of messages, taking care of Alpha-renaming of the variables. It is a theorem 1 checker:if some 
-- protocol is format type safe

module FormatTypeSafe (StateType, subTerms, msgPattern,protocolMP, protocolNVST, alphaRenaming,isNonVarSubTerm,checkUnify,theoremNum1,zipName) where
import Message
import Unification
import Control.Monad.State
import qualified Data.Map as Map


-- | Data type for state monad. Type of the value of the state is the same. Function of the int is to keep counter and do alpha renaming
type StateType = (Map.Map (Message,String) MessageType, Int)


-- | Function that makes non variable subterms of single message and its type, doing alpha-renaming 
subTerms :: ((Message,String), MessageType) -> State StateType StateType

subTerms  ((Atom(Ident x),protocolName), msgtype) = do
      (state, counter) <- get
      if isConstant x
              then put ( Map.insert  ((alphaRenaming (Atom (Ident x)) counter),protocolName) msgtype $ state ,counter+1)
              else put (state,counter)
      get

subTerms ((Atom(Mapping id t),protocolName), keyType)  = do
      (state, counter) <- get
      put ( Map.insert ((alphaRenaming (Atom(Mapping id t)) counter),protocolName) keyType $ state , counter+1)
      get

subTerms ((Asycrypt k x,protocolName), Crypt key msg) = do
      (state, counter) <- get
      put ( Map.insert ((alphaRenaming (Asycrypt k x) counter),protocolName) (Crypt key msg) $ state, counter+1)
      if isNonVarSubTerm k
              then do subTerms ((k,protocolName),key)
                      subTerms ((x,protocolName),msg)
              else subTerms ((x,protocolName),msg)
      
subTerms ((Symcrypt k x,protocolName), Scrypt key msg) = do
      (state, counter) <- get
      put ( Map.insert ((alphaRenaming (Symcrypt k x) counter),protocolName) (Scrypt key msg) $ state, counter+1)
      subTerms ((k,protocolName),key)
      subTerms ((x,protocolName),msg)

subTerms ((Concat x,protocolName), Conc t) = do
      (state, counter) <- get
      put (Map.insert ((alphaRenaming (Concat x) counter),protocolName) (Conc t) $ state, counter+1)
      mapM subTerms (zip (zipName x protocolName) t) 
      get

subTerms ((Hash id x,protocolName), HashT t) = do
      (state, counter) <- get
      put (Map.insert ((alphaRenaming (Hash id x) counter),protocolName) (HashT t) $ state, counter+1)
      subTerms ((x,protocolName),t)

-- | function operates on the single message to do variable renamings.

msgPattern :: ((Message,String), MessageType) -> State StateType StateType


msgPattern  ((Atom(Ident x),protocolName), msgtype) = do
      (state, counter) <- get
      put ( Map.insert  ((alphaRenaming (Atom (Ident x)) counter),protocolName) msgtype $ state ,counter+1)
      get

msgPattern ((Atom(Mapping id t),protocolName), keyType)  = do
      (state, counter) <- get
      put ( Map.insert ((alphaRenaming (Atom(Mapping id t)) counter),protocolName) keyType $ state , counter+1)
      get


msgPattern ((Asycrypt k x,protocolName), Crypt key msg) = do
      (state, counter) <- get
      put ( Map.insert ((alphaRenaming (Asycrypt k x) counter),protocolName) (Crypt key msg) $ state, counter+1)
      get

msgPattern ((Symcrypt k x,protocolName), Scrypt key msg) = do
      (state, counter) <- get
      put ( Map.insert ((alphaRenaming (Symcrypt k x) counter),protocolName) (Scrypt key msg) $ state, counter+1)
      get

msgPattern ((Concat x,protocolName), Conc t) = do
      (state, counter) <- get
      put (Map.insert ((alphaRenaming (Concat x) counter),protocolName) (Conc t) $ state, counter+1)
      get

msgPattern ((Hash id x,protocolName), HashT t) = do
      (state, counter) <- get
      put (Map.insert ((alphaRenaming (Hash id x) counter),protocolName) (HashT t) $ state, counter+1)
      get
-------------------- helpful function
--
--

zipName :: [Message] -> String -> [(Message,String)]
zipName [] name = []
zipName (x:xs) name = (x,name):(zipName xs name)

-- | function is getting message pattern. It uses msgPattern function

protocolMP :: [((Message,String), MessageType)] -> (Map.Map (Message,String) MessageType, Int) -> (Map.Map (Message,String) MessageType, Int) 
protocolMP [] initState = initState -- fst initstate, then function only gives back one thing
protocolMP ((msg,msgtype):xs) initState = let newInitState = evalState (msgPattern (msg,msgtype)) initState -- here msg is msg+protocol name
                                            in protocolMP xs newInitState

-- | function counts non variable subterms of one protocol
protocolNVST :: [((Message,String), MessageType)] -> (Map.Map (Message,String) MessageType, Int) -> (Map.Map (Message,String) MessageType, Int) 
protocolNVST [] initState = initState 
protocolNVST ((msg,msgtype):xs) initState = let newInitState = evalState (subTerms (msg,msgtype)) initState -- msg = msg+protocol name
                                            in protocolNVST xs newInitState

-- | alpha renaming function
alphaRenaming :: Message -> Int -> Message
alphaRenaming (Atom (Ident t)) num | (isVariable t) = Atom (Ident (t++show num))
                                   | otherwise = (Atom (Ident t))
alphaRenaming (Atom (Mapping id t)) num = Atom (renameAtomic (Mapping id t) num)
alphaRenaming (Symcrypt key t) num = Symcrypt (alphaRenaming key num) (alphaRenaming t num)
alphaRenaming (Asycrypt key t) num = Asycrypt (alphaRenaming key num) (alphaRenaming t num)
alphaRenaming (Hash id t) num = Hash id (alphaRenaming t num)
alphaRenaming (Concat t) num = Concat (map (\ x -> alphaRenaming x num) t)

-- | aux function for alphaRenaming
renameAtomic :: Atomic ->Int -> Atomic -- just because the structure of Mapping
renameAtomic (Ident x) num = Ident (x++ show num)
renameAtomic (Mapping id t) num = Mapping id (map (\x -> renameAtomic x num) t )

-- helpful function to see nonvarsubterms
isNonVarSubTerm:: Message -> Bool -- filter for Map.Map --- additional stuff
isNonVarSubTerm (Atom (Mapping _ _)) = False
isNonVarSubTerm (Atom (Ident t)) = isConstant t
isNonVarSubTerm _ = True


-- | unification checker between subterms of message pattern. theorem 1 is violated if unification is possible between subterms of different types
checkUnify :: Map.Map (Message,String) MessageType -> [(Message,String)] -> Bool  --- main check for theorem
checkUnify table [] = True 
checkUnify table (x:arrayMsgtail) = 
                    (foldl (\acc a -> if ((unify[(fst x,fst a)] /= Nothing) && (table Map.! x /= table Map.!a)) 
                                        then do error ("\n Protocol "++show (snd x)++" is not format-type safe! \n Unification is possible between terms: "++show (fst x)++ " and " ++ show (fst a)) else acc) True arrayMsgtail)
                    && checkUnify table arrayMsgtail

-- | first condition of theorem 1 (format type safe)
cond1Theorem1 :: [((Message,String), MessageType)] -> Bool -- MP(P) intersection with variable set is empty?
cond1Theorem1 [] = True
cond1Theorem1 array = foldl (\acc a -> if (isNonVarSubTerm (fst (fst a))) 
                                        then acc 
                                        else do error ("\n Protocol"++show(snd(fst a))++". Element of the message pattern: "++show(fst(fst a))++" is variable!!!")) True array 

-- | central function for checking theorem 1: format safe protocol (page 13 in extended version)
theoremNum1 :: [((Message,String), MessageType)] -> Bool  
theoremNum1 array = let nvst = protocolNVST array (Map.empty,0)
                    in 
                        if (cond1Theorem1 array) && (checkUnify (fst nvst) (Map.keys (fst nvst)))
                                then True
                                else error ("Theorem 1 is violated!!!") -- never here


transformFormat :: ([(Message,MessageType)],String) -> [((Message,String), MessageType)]
transformFormat ([], protocolName) = []
transformFormat (((msg,msgType):array),protocolName) = ((msg,protocolName),msgType): (transformFormat (array, protocolName))



------------------------------------- testing the module


