{-
 -
 - APCC
 - (C) Ivana Kojovic, DTU 2012
 - All Rights Reserved.
 -
 - Unification by the book "Message rewriting and all that"

27th April, added Show typeclass

-}

-- | Module that models @Message@ data type used in analysis

module Message (Id, Message(..), BasicType(..), MessageType(..), isConstant, isVarTerm, isVariable, isAtom, Atomic (..),isMapping) 
     where 

type Id = String -- | type for all the symbols (constants, vars, functions...)


data Atomic = Ident Id -- ^ atomic part of message to represent identifier 
          | Mapping Id [Atomic] -- ^ e.g. pk : agent -> pubkey - atomic term
          deriving (Eq,Ord) -- ^ Ord is necessary for being a key in Map.map data type
          
-- | Main data type
data Message = Atom Atomic -- ^ it can be atomic message (term) 
              | Concat [Message] -- ^ concatenation of messages
              | Asycrypt Message Message  -- ^ asymmetric encryption, first message is a key and the other one is a message to be encrypted
              | Symcrypt Message Message  -- ^ symmetric encryption, first message is a key and the other one is a message to be encrypted
              | Hash Id Message -- ^ hashing (with its name -id) of the message
              deriving (Eq,Ord) 


-- | Typing system of the messages, "label" for all the terms in the analysis
data BasicType = Agent 
                | Nonce
                | Symkey
                | Pubkey
                | Privkey
                | Tag
                deriving (Eq,Ord)

-- MessageType  - composed types, depending of the message structure

data MessageType = Basic BasicType
                | Conc [MessageType] -- ^ concatenation type
                | Crypt MessageType MessageType -- ^ type of the symmetric encryption message
                | Scrypt MessageType MessageType -- ^ type of the asymmetric encryption message
                | HashT MessageType -- ^ all the hash functions are having HashT type
                deriving (Eq,Ord)


-- | pretty print for BasicType
instance Show BasicType where
    show Agent = "agent"
    show Nonce = "nonce"
    show Symkey = "symkey"
    show Pubkey = "pubkey"
    show Privkey = "privkey"
    show Tag = "tag"

-- | pretty print for MessageType

instance Show MessageType where
    show (Basic msgt) = show msgt 
    show (Conc msgs) = show msgs
    show (Crypt keyt msgt) = "crypt("++show keyt++","++show msgt++")"
    show (Scrypt keyt msgt) = "scrypt("++show keyt++","++show msgt++")"
    show (HashT msgs) = "hash("++show msgs++")"



-- | pretty print for Message

instance Show Message where  
    show (Atom (Ident id)) = id  
    show (Atom (Mapping id t)) = id++"("++ show t++")"
    show (Concat t) = show t
    show (Asycrypt k m) = "{"++ show m ++"}"++ show k
    show (Symcrypt k m) = "|" ++ show m ++ "|" ++ show k
    show (Hash id t) = id ++ show t

instance Show Atomic where
   show (Ident t) = t
   show (Mapping id t) = id++"("++ show t++")"

-- | checking whether some string is a constant
isConstant :: String -> Bool
isConstant x = elem (head  x) ['a'..'z']

-- | checking whether some term represents is a variable

isVarTerm :: Message -> Bool -- 
isVarTerm (Atom (Ident id)) = isVariable id
isVarTerm _ = False

-- | checking whether some message is a mapping
isMapping :: Message -> Bool 
isMapping (Atom (Mapping _ _)) = True
isMapping _ = False

-- | checking whether some string is a variable
isVariable :: String -> Bool  
isVariable x = elem (head x) ['A'..'Z']


-- | checking whether some message is atomic
isAtom :: Message -> Bool 
isAtom (Atom _) = True      
isAtom _ = False



isNonce :: Message -> Bool ---  not used, still
isNonce (Atom (Ident x))= elem (head x) ['N'] 

isAgent:: Message -> Bool -- not used still
isAgent (Atom (Ident x)) = elem (head x) (['A'..'F'] ++ ['a'..'f'])  ----  



checkTypeOfMsg :: Message -> MessageType -- not used
checkTypeOfMsg (Atom (Ident x)) | isAgent (Atom (Ident x)) = Basic Agent 
                                | isNonce (Atom (Ident x)) = Basic Nonce 
                                | otherwise = Basic Tag

-------------------------------------------------------------------------------
--

