{-
 -
 - APCC
 - (C) Ivana Kojovic, DTU 2012
 - All Rights Reserved.
 -
 - Unification by the book "Message rewriting and all that"
 -}

module Main where
import AnBParser
import Unification
import Lexer
import Ast
import Msg 
import qualified Data.List as List
import AnBOnP
import System.IO
import qualified Data.Map as Map
import qualified Message as M
import System.Environment(getArgs)
import Checker
import System.Directory(getDirectoryContents)


checkName :: Msg -> Ident -- give back name
checkName (Atom id) = id


protocolName :: Protocol -> String  -- name of the protocol
protocolName (name,_,_,_,_,_,_,_,_) = name

protocolActions :: Protocol -> Actions --  on 7th place, list all the actions
protocolActions (_,_,_,_,_,_,actions,_,_) = actions  -- it is array [action]

protocolPrivate :: Protocol -> [Msg]
protocolPrivate (_,_,_,_,private,_,_,_,_) = private

protocolPublic :: Protocol -> [Msg]
protocolPublic (_,_,_,public,_,_,_,_,_) = public


msgfromAction :: Action -> Msg -- give msg from action
msgfromAction (_,msg,_,_) = msg

extractActions :: Protocol -> [Msg]
extractActions protocol = map msgfromAction (protocolActions protocol)

protocolTypes :: Protocol -> Types
protocolTypes (_,types,_,_,_,_,_,_,_) = types

isConcat :: M.Message -> Bool
isConcat (M.Concat whatever) = True
isConcat _ = False




msgTrans :: Msg -> M.Message -- from Msg to Message
msgTrans (Atom id) = M.Atom (M.Ident id) -- agent/nonce

msgTrans (Comp Apply [Atom "sk", Comp Cat array]) = M.Atom ((M.Mapping "sk") [ M.Ident (checkName (head array)),M.Ident (checkName (last array))]) -- shared key
msgTrans (Comp Apply array) | (checkName (head array) == "pk") = M.Atom ((M.Mapping "pk") [ M.Ident (checkName (last array))]) -- publickey
                            | otherwise = M.Hash (checkName(head array)) (msgTrans (last array)) -- hash functions

msgTrans (Comp Inv [Comp Apply array]) = M.Atom ((M.Mapping "inv") [(M.Mapping (checkName (head array))) [M.Ident (checkName (last array))]]) --private key
msgTrans (Comp Cat array) = M.Concat (map msgTrans array) -- concatenation
msgTrans (Comp Scrypt array) = M.Symcrypt (msgTrans (head(array))) (msgTrans (last(array))) --- sym encryption
msgTrans (Comp Crypt array) = M.Asycrypt (msgTrans (head(array))) (msgTrans (last(array))) --- asym. encryption

solveConcat :: M.Message -> M.Message
solveConcat (M.Atom (M.Ident id)) = (M.Atom (M.Ident id))
solveConcat (M.Atom (M.Mapping a array)) = (M.Atom (M.Mapping a array))
solveConcat (M.Symcrypt k m) = M.Symcrypt (solveConcat k) (solveConcat m)
solveConcat (M.Asycrypt k m) = M.Asycrypt (solveConcat k) (solveConcat m)
solveConcat (M.Hash id m) = M.Hash id (solveConcat m)
solveConcat (M.Concat array) = M.Concat (map solveConcat (takeList(M.Concat array)))

takeList :: M.Message -> [M.Message]
takeList (M.Concat a) = if (isConcat (last a))
                                    then [head a] ++ takeList (last a)
                                    else a
takeList msg = [msg]

-- main function for transformation, calling solveConcat and msgTrans
msgToMessage :: Msg -> M.Message
msgToMessage msg = solveConcat(msgTrans msg)

isAgent :: Types -> String -> Bool
isAgent [] id = False
isAgent typesArray id = let aux = head (typesArray) 
                                    in 
                                        if (fst(aux) == (Agent False False)) && (elem id (snd(aux)))
                                                then True
                                                else isAgent (tail typesArray) id

isNonce :: Types -> String -> Bool
isNonce [] id = False
isNonce typesArray id = let aux = head (typesArray) 
                                    in 
                                        if (fst(aux) == Number) && (elem id (snd(aux)))
                                                then True
                                                else isNonce (tail typesArray) id


isSkey :: Types -> String -> Bool
isSkey [] id = False
isSkey typesArray id = let aux = head (typesArray) 
                                    in 
                                        if (fst(aux) == SymmetricKey) && (elem id (snd(aux)))
                                                then True
                                                else isSkey (tail typesArray) id

isPkey :: Types -> String -> Bool
isPkey [] id = False
isPkey typesArray id = let aux = head (typesArray) 
                                    in 
                                        if (fst(aux) == PublicKey) && (elem id (snd(aux)))
                                                then True
                                                else isPkey (tail typesArray) id

isMappingF :: Types -> String -> Bool
isMappingF [] id = False
isMappingF types id = let aux = head (types)
                            in if (elem id (snd aux)) && ((fst aux) == Mapping)
                                    then True
                                    else isMappingF (tail types) id


identType :: Types -> String -> Type
identType [] id = error ("Function is not defined!")
identType types id = let aux = head (types)
                            in if (elem id (snd aux))
                                    then fst (aux)
                                    else identType (tail types) id


assignType :: M.Message -> Protocol -> M.MessageType
assignType (M.Atom (M.Ident id)) p = if (isAgent (protocolTypes p) id)
                                        then M.Basic M.Agent
                                        else if (isSkey (protocolTypes p) id)
                                                then M.Basic M.Symkey
                                                else if (isPkey (protocolTypes p) id)
                                                        then M.Basic M.Pubkey
                                                        else if (isNonce (protocolTypes p) id)
                                                                then if (List.isPrefixOf "tag" id)
                                                                        then M.Basic M.Tag
                                                                        else M.Basic M.Nonce
                                                                else error ("Identifier "++show id++" is not typed!")


assignType (M.Atom (M.Mapping id array)) p= if (id == "pk") && (isMappingF (protocolTypes p) id)
                                                then M.Basic M.Pubkey
                                                else
                                                    if (id == "sk") && (isMappingF (protocolTypes p) id)
                                                            then M.Basic M.Symkey
                                                            else 
                                                                    if (id == "inv")
                                                                            then M.Basic M.Privkey
                                                                            else error ("Undefined mapping" ++ id)

assignType (M.Asycrypt key msg) p = M.Crypt (assignType key p) (assignType msg p)
assignType (M.Symcrypt key msg) p = M.Scrypt (assignType key p) (assignType msg p)
assignType (M.Hash id msg) p = if (identType (protocolTypes p) id) == Function
                                    then M.HashT (assignType msg p)
                                    else error "Please, define name of the function!"

assignType (M.Concat array) p = M.Conc (map (\x -> assignType x p) array)


processProtocol :: Protocol -> ([(M.Message, M.MessageType)],String)
processProtocol protocol = let actions = extractActions protocol -- [Action]
                               name = protocolName protocol
                               msgs = map msgToMessage actions -- [Message] of the protocols
                               typeMsgs = map (\x -> assignType x protocol) msgs
                            in (zip msgs typeMsgs,name) 

wrapPrivateForParallel :: Protocol -> [((M.Message,String),M.MessageType)] -- ([(M.Message, M.MessageType)],String)
wrapPrivateForParallel protocol = let ofmcMessages = protocolPrivate protocol -- [Msg]
                                      name = protocolName protocol
                                      msgs = map msgToMessage ofmcMessages -- [Message] of the protocols
                                      typeMsgs = map (\x -> assignType x protocol) msgs
                                      wrapped = (zip msgs typeMsgs,name)
                                   in putNameInsideMessages wrapped

wrapPublicForParallel :: Protocol -> [((M.Message,String),M.MessageType)] -- ([(M.Message, M.MessageType)],String)
wrapPublicForParallel protocol = let ofmcMessages = protocolPublic protocol --
                                     name = protocolName protocol
                                     msgs = map msgToMessage ofmcMessages -- [Message] of the protocols
                                     typeMsgs = map (\x -> assignType x protocol) msgs
                                     wrapped = (zip msgs typeMsgs,name)
                                  in putNameInsideMessages wrapped

unifyParij :: Map.Map (M.Message,String) M.MessageType -> Map.Map (M.Message,String) M.MessageType -> Bool
unifyParij x y = if ((not (Map.null x)) && (not (Map.null y)))
                                  then let new = head(Map.toList x)
                                        in (unifyAuxPar new y (Map.keys y)) || (unifyParij (Map.delete (fst new) x) y)
                                  else False




unifyAuxPar :: ((M.Message,String),M.MessageType) -> Map.Map (M.Message,String) M.MessageType -> [(M.Message,String)] -> Bool
unifyAuxPar (msg,msgType) table array =  foldl (\acc a -> if ((unify[(fst msg,fst a)] /= Nothing) && (msgType == table Map.!a)) -- msg is message+protocolname!!
                                                            then error ("\n Parallel composition is not possible!  \n Unification between terms: "++ show (fst msg) ++" in protocol "++ show (snd msg) ++ " and " ++ show (fst a) ++ " in protocol " ++ show (snd a)) 
                                                            else acc) False array


parallelChecker :: [Protocol] -> Bool
parallelChecker [] = False
parallelChecker (x:arrayTail) = (foldl (\acc a -> if (parallelCheckerij x a) then True else acc ) False arrayTail) || parallelChecker arrayTail

parallelCheckerij :: Protocol -> Protocol -> Bool
parallelCheckerij p1 p2 = let p1private = wrapPrivateForParallel p1
                              p1public = wrapPublicForParallel p1
                              p2private = wrapPrivateForParallel p2
                              p2public = wrapPublicForParallel p2
                            in (unifyParij (Map.fromList p1private) (Map.fromList p2public)) || (unifyParij (Map.fromList p1public) (Map.fromList p2private))
                                
main = do

    xs <- getArgs
    if null xs then do
        putStrLn "You entered no arguments"
     else do
        putStrLn ("You entered " ++ show xs)
        --file <- mapM readFile xs
        dir <- getDirectoryContents (head xs) >>= return . filter (List.isSuffixOf ".AnB") -- added afterwards
        file <- mapM readFile (map (\x -> ((head xs)++"/"++x)) dir) -- also!

        let protocols = map anbparser (map alexScanTokens file)
            allProtocols = map processProtocol protocols -- [[Message,MessageType]], ready for checker!
        putStrLn (" *******  APCC ********** ")
        putStrLn ("Are conditions for vertical composition being violated? :  "++ show (disjointCheck allProtocols))
        putStrLn ("Are parallel composition conditions being violeted? :    " ++ show (parallelChecker protocols))
   




